from abc import ABCMeta
from typing import Type, List, AnyStr, Tuple

from django.db import transaction
from django.db.models import Model


class BaseCommentMigration(metaclass=ABCMeta):
    atomic = True

    def __init__(self, connection, model: Type[Model], collect_sql=False):
        self.connection = connection
        self.model = model
        self.collect_sql = collect_sql
        if self.collect_sql:
            self.collected_sql = []

    def comments_sql(self) -> List[Tuple[AnyStr, List[AnyStr]]]:
        pass

    def migrate_comments_to_database(self):
        pass

    def quote_name(self, name):
        return self.connection.ops.quote_name(name)

    def execute(self):
        if self.atomic:
            with transaction.atomic():
                self.execute_sql()
        else:
            self.connection.needs_rollback = False
            self.execute_sql()

    def execute_sql(self):
        comments_sql = self.comments_sql()
        if comments_sql:
            with self.connection.cursor() as cursor:
                for comment_sql, params in comments_sql:
                    cursor.execute(comment_sql, params)



from django_comment_migrate.backends.base import BaseCommentMigration
from django_comment_migrate.utils import get_field_comment


class CommentMigration(BaseCommentMigration):
    atomic = False
    sql_alter_column = "ALTER TABLE %(table)s %(changes)s"
    sql_alter_column_comment_null = "MODIFY COLUMN %(column)s %(type)s NULL" \
                                    " COMMENT %(comment)s"
    sql_alter_column_comment_not_null = "MODIFY COLUMN %(column)s %(type)s " \
                                        "NOT NULL COMMENT %(comment)s"

    def comments_sql(self):
        db_table = self.model._meta.db_table
        changes = []
        params = []
        for field in self.model._meta.fields:
            comment = get_field_comment(field)
            if comment:
                db_parameters = field.db_parameters(connection=self.connection)
                sql = self.sql_alter_column_comment_null if field.null \
                    else self.sql_alter_column_comment_not_null
                changes.append(sql % {
                    "column": self.quote_name(field.column),
                    "type": db_parameters['type'],
                    "comment": "%s"
                })
                params.append(comment)
        if changes:
            return [
                (
                    self.sql_alter_column % {
                        "table": db_table,
                        "changes": ",".join(changes)
                    },
                    params
                )
            ]



import sys

from django.apps import apps
from django.core.management import BaseCommand
from django.db import DEFAULT_DB_ALIAS, router

from django_comment_migrate.db_comments import migrate_app_models_help_text_to_database


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            '--database', default=DEFAULT_DB_ALIAS,
            help='Nominates a database to migrate.'
                 ' Defaults to the "default" database.',
        )
        parser.add_argument(
            'app_label', nargs='?',
            help='App labels of applications to limit the migrate comment'
        )

    def handle(self, *args, **options):
        using = options['database']
        app_label = options['app_label']
        if app_label:
            app_configs = self.filter_valid_app_configs([app_label])
        else:
            app_configs = self.load_app_configs(using)

        for app_config in app_configs:
            migrate_app_models_help_text_to_database(app_config.get_models(), using)
            self.stdout.write(self.style.SUCCESS(
                f"migrate app {app_config.label} successful"))

    def load_app_configs(self, using):
        migrated_apps = set()
        for app_config in apps.get_app_configs():
            app_label = app_config.label
            if router.allow_migrate(using, app_label):
                migrated_apps.add(app_config)
            else:
                self.stdout.write(f"app {app_label}  not allow migration")
        return migrated_apps

    def filter_valid_app_configs(self, app_names):
        has_bad_names = False
        migrated_apps = set()
        for app_name in app_names:
            try:
                migrated_apps.add(apps.get_app_config(app_name))
            except LookupError as error:
                self.stderr.write(error)
                has_bad_names = True
        if has_bad_names:
            # 2 代表误用了命令
            sys.exit(2)
        return migrated_apps



from django.apps import AppConfig, apps
from django.db import DEFAULT_DB_ALIAS
from django.db.models.signals import post_migrate

from django_comment_migrate.db_comments import migrate_app_models_help_text_to_database
from django_comment_migrate.utils import get_migrations_app_models


def handle_post_migrate(app_config, using=DEFAULT_DB_ALIAS, **kwargs):
    migrations = (migration for migration, _ in kwargs.get('plan', []))
    app_models = get_migrations_app_models(migrations, apps, using)
    migrate_app_models_help_text_to_database(app_models, using)


class DjangoCommentMigrationConfig(AppConfig):
    name = "django_comment_migrate"

    def ready(self):
        post_migrate.connect(handle_post_migrate)



from django.db import router, DEFAULT_DB_ALIAS
from django.db.migrations import Migration
from django.db.models import Field


def get_field_comment(field: Field):
    if field.help_text:
        return str(field.help_text)


def get_migrations_app_models(migrations: [Migration], apps, using=DEFAULT_DB_ALIAS):
    models = set()
    for migration in migrations:
        if not isinstance(migration, Migration):
            continue
        app_label = migration.app_label
        if not router.allow_migrate(using, app_label):
            return
        operations = getattr(migration, 'operations', [])
        for operation in operations:
            model_name = getattr(operation, 'model_name', None) or getattr(operation, 'name', None)
            if model_name is None:
                continue
            model = apps.get_model(app_label, model_name=model_name)
            models.add(model)
    return models
